/*
   FILE:
						       	
	rtl_tasks.c
								       	
   DESCRIPTION:							       	
	RT-Linux side of a sample application using the POSIX trace
	support.

	See README file for details.
					       				
   Copyright (C) 2001
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        September 2001

*/

#include <rtl.h>
#include <time.h>
#include <pthread.h>
#include <rtl_sched.h>
#include <trace.h>
#include <rtl_fifo.h>
#include <unistd.h>


static int delay = 15;
static int fifo = 10;

MODULE_PARM(delay,"i");
MODULE_PARM(fifo,"i");


//#define _DEBUG

static trace_id_t       trid;

static int              fifo_desc;

static pthread_mutex_t mutex, mutex2;
static pthread_t       task2, task3, task4, task5, task6, task7, task8;
static pthread_t       task9, task10, task11, task12, task13, task14, task15;
static pthread_t       task_shutdown;



void *just_execute(void *task_period) {
  int       j;
  long      period;
  struct timespec next_period;

  period = (long) task_period;

  clock_gettime(CLOCK_MONOTONIC, &next_period);

  while (1) {
    timespec_add_ns(&next_period, period);
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);

    for (j=0; j<period/800; j++);

    pthread_mutex_lock(&mutex);
      for (j=0; j<period/600; j++);
    pthread_mutex_unlock(&mutex);

    for (j=0; j<period/2000; j++);

  }

  pthread_exit(0);
  return (void *) 0;
}


void *just_execute_long(void *task_period) {
  int       j;
  long      period;
  struct timespec next_period;

  period = (long) task_period;

  clock_gettime(CLOCK_MONOTONIC, &next_period);

  while (1) {
    timespec_add_ns(&next_period, period);
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);

    for (j=0; j<period/5000; j++);
    pthread_mutex_lock(&mutex2);
    for (j=0; j<period/1000; j++);
    pthread_mutex_unlock(&mutex2);
    for (j=0; j<period/10000; j++);
  }

  pthread_exit(0);
  return (void *) 0;
}


 
void *just_execute_very_long(void *seconds_period) {
  int       j;
  long      seconds;
  struct timespec next_period;

  seconds = (long) seconds_period;

  clock_gettime(CLOCK_MONOTONIC, &next_period);

  while (1) {
    next_period.tv_sec += seconds;
    clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &next_period, NULL);

    for (j=0; j<(seconds*10000); j++);
  }

  pthread_exit(0);
  return (void *) 0;
}






void *shutdown_task(void *dummy) {
  struct timespec wakeup_time;

  clock_gettime(CLOCK_MONOTONIC, &wakeup_time);
  wakeup_time.tv_sec += delay;
  clock_nanosleep(CLOCK_MONOTONIC, TIMER_ABSTIME, &wakeup_time, NULL);
#ifdef _DEBUG
  rtl_printf("(shutdown_task): Waking up to shutdown the stream %d\n", trid);
#endif
  posix_trace_shutdown(trid);

  return (void *) 0;
}





int init_module(void) {

  trace_attr_t         attr;
  pthread_mutexattr_t  mutattr;
  pthread_attr_t       thattr;
  int                  error;
  char                 fifo_name[16];
  struct sched_param   s_param;

  // Create mutexes
  pthread_mutexattr_init(&mutattr);
  if (pthread_mutexattr_setprotocol(&mutattr, PTHREAD_PRIO_PROTECT) != 0) {
    rtl_printf("No prio protect\n");
    return 1;
  }
  pthread_mutexattr_setprioceiling(&mutattr,
				   sched_get_priority_min(SCHED_FIFO) + 16);
  pthread_mutex_init(&mutex, &mutattr);
  pthread_mutexattr_setprioceiling(&mutattr,
				   sched_get_priority_min(SCHED_FIFO) + 11);
  pthread_mutex_init(&mutex2, &mutattr);

  // Create and set the trace attribute:
  error = posix_trace_attr_init(&attr);
  error = posix_trace_attr_setstreamfullpolicy (&attr, POSIX_TRACE_UNTIL_FULL);
  error = posix_trace_attr_setmaxdatasize(&attr, 16);
  error = posix_trace_attr_setstreamsize(&attr, 20000000);


  // Create/open the fifo
  sprintf(fifo_name, "/dev/rtf%d", fifo);
#ifdef _DEBUG
  rtl_printf("(init_module) Creating fifo No. %d (%s)\n", fifo, fifo_name);
#endif
  rtf_create(fifo, 100000);

#ifdef _DEBUG
  rtl_printf("(init_module) Opening fifo No. %d (%s)\n", fifo, fifo_name);
#endif
  fifo_desc = open(fifo_name, O_NONBLOCK);
  if (fifo_desc < 0) {
    rtl_printf("(init_module): Cannot open %s for writing\n", fifo_name);
    return -1;
  }

  // Create the stream:
#ifdef _DEBUG
  rtl_printf("(init_module) Creating stream witlog %d\n", fifo_desc);
#endif
  error = posix_trace_create_withlog(0, &attr, fifo_desc, &trid); 
  if (error) {
    rtl_printf("(init_module): Cannot create trace with log %d. Error=%d\n", 
	       fifo_desc, error);
    return -1;
  }

   // Start tracing:
#ifdef _DEBUG
  rtl_printf("(init_module) Starting trace stream %d\n", (int) trid);
#endif
  error = posix_trace_start(trid);
  if (error) {
    rtl_printf("(init_module): Cannot start trace %d. Error=%d\n", 
	       (int) trid, error);
    return -1;
  }


#ifdef _DEBUG
  rtl_printf("(init_module) Creating tasks\n");
#endif

  pthread_attr_init (&thattr); 
  pthread_create (&task_shutdown, &thattr, shutdown_task, NULL); 
 
  // This one awakes each 15 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 14;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task2, &thattr, just_execute, (void *) 15000000); 

  // This one awakes each 30 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 13;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task3, &thattr, just_execute, (void *)  30000000); 

  // This one awakes each 45 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 12;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task4, &thattr, just_execute, (void *) 45000000); 

  // This one awakes each 80 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 11;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task5, &thattr, just_execute, (void *) 80000000); 

  // This one awakes each 150 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 10;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task6, &thattr, just_execute_long, (void *) 150000000); 
 

  // This one awakes each 230 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 9;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task7, &thattr, just_execute_long, (void *) 230000000); 


  // This one awakes each 500 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 8;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task8, &thattr, just_execute_long, (void *) 500000000); 

  // This one awakes each 800 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 7;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task9, &thattr, just_execute_long, 
		  (void *) 800000000); 

  // This one awakes each 900 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 6;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task10, &thattr, just_execute_long, 
		  (void *) 900000000); 


  // This one awakes each 1000 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 5;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task11, &thattr, just_execute_very_long, 
		  (void *) 1); 


  // This one awakes each 2000 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 4;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task12, &thattr, just_execute_very_long, 
		  (void *) 2); 

  // This one awakes each 3000 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 3;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task13, &thattr, just_execute_very_long, 
		  (void *) 3); 

  // This one awakes each 4000 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 2;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task14, &thattr, just_execute_very_long, 
		  (void *) 4); 

  // This one awakes each 5000 msec:
  s_param.sched_priority = sched_get_priority_min(SCHED_FIFO) + 1;
  pthread_attr_setschedparam(&thattr, &s_param);
  pthread_create (&task15, &thattr, just_execute_very_long, 
		  (void *) 5); 



  return 0;
}



void cleanup_module(void) {

  rtl_printf("rtl_tasks: CLEANUP!!!\n");

  pthread_cancel(task_shutdown); 
  pthread_join(task_shutdown,NULL); 

  pthread_cancel(task2); 
  pthread_cancel(task3); 
  pthread_cancel(task4); 
  pthread_cancel(task5); 
  pthread_cancel(task6); 
  pthread_cancel(task7); 
  pthread_cancel(task8); 
  pthread_cancel(task9); 
  pthread_cancel(task10); 
  pthread_cancel(task11); 
  pthread_cancel(task12); 
  pthread_cancel(task13); 
  pthread_cancel(task14); 
  pthread_cancel(task15); 

  pthread_join(task2,NULL); 
  pthread_join(task3,NULL); 
  pthread_join(task4,NULL); 
  pthread_join(task5,NULL); 
  pthread_join(task6,NULL); 
  pthread_join(task7,NULL); 
  pthread_join(task8,NULL); 
  pthread_join(task9,NULL); 
  pthread_join(task10,NULL); 
  pthread_join(task11,NULL); 
  pthread_join(task12,NULL); 
  pthread_join(task13,NULL); 
  pthread_join(task14,NULL); 
  pthread_join(task15,NULL); 

  close(fifo_desc);
  rtf_destroy(fifo);
}
