/*
   FILE:							       	
	trace/trace_options.h							       
								       	
   DESCRIPTION:							       	
	Establishes which tracing options are currently supported. 
					       				
   Copyright (C) 2002
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
       December 2003

*/

#ifndef __TRACE_POSIX_TRACE_OPTIONS_H
#define __TRACE_POSIX_TRACE_OPTIONS_H

/*
  CHANGELOG:

        Version 2.0 (December 2003)

	Symbolic constants expressing the trace options supported are
	now changed to others reflecting the *units of functionality*
	supported. Some of them are analogous to former trace options,
	but normally units are subsets of options.


        Version 1 (April 2002)

	Symbolic constants as defined in the standard (section "2.9
        SYMBOLIC CONSTANTS" in the P1003.1q amendment: _POSIX_TRACE,
        _POSIX_TRACE_EVENT_FILTER, _POSIX_TRACE_LOG and
        _POSIX_TRACE_INHERITANCE
*/


/* 
   1.3 CONFORMANCE     
   (Values are defined according to Section "2.9 SYMBOLIC CONSTANTS")

   NOTE:  P1003.1q was finally approved on Sep, 21st 2000, so...
*/



// Options selected when compiling RT-Linux are placed in rtl_config.h
#include <rtl_conf.h>


//
// (1) Units of functionality:
//

#ifdef CONFIG_OC_PTRACE_CORE
#define _POSIX_TRACE_CORE                  200312L
#endif

#ifdef CONFIG_OC_PTRACE_ONLINE
#define _POSIX_TRACE_ONLINE                200312L
#endif

#ifdef CONFIG_OC_PTRACE_TOLOG 
#define _POSIX_TRACE_TOLOG                 200312L
#endif

#ifdef CONFIG_OC_PTRACE_UEVENTS
#define _POSIX_TRACE_UEVENTS               200312L
#endif

#ifdef CONFIG_OC_PTRACE_FILTER
#define _POSIX_TRACE_EVENT_FILTER          200312L
#endif

#ifdef CONFIG_OC_PTRACE_FILTER
#define _POSIX_TRACE_FILTER                200312L
#endif

#ifdef CONFIG_OC_PTRACE_INHERITANCE
#define _POSIX_TRACE_INHERITANCE           200312L
#endif

#ifdef CONFIG_OC_PTRACE_LOGRET     
#define _POSIX_TRACE_LOGRET                200312L
#endif


//
// (2) Maximums and limits
//

#ifdef CONFIG_OC_PTRACE_SINGLESTREAM
#define _POSIX_TRACE_SINGLESTREAM
#define _POSIX_TRACE_STREAMS               1
#elif CONFIG_OC_PTRACE_STREAMS    
#define _POSIX_TRACE_STREAMS               (CONFIG_OC_PTRACE_STREAMS <= 8 \
                                            && CONFIG_OC_PTRACE_STREAMS > 0 ? \
                                           CONFIG_OC_PTRACE_STREAMS : \
                                           8)
#else
#define _POSIX_TRACE_STREAMS               8
#endif


#ifdef CONFIG_OC_PTRACE_STSIZE 
#define _POSIX_TRACE_STREAM_SIZE           CONFIG_OC_PTRACE_STSIZE   
#endif


#ifdef CONFIG_OC_PTRACE_LIMEVSIZE 
#define _POSIX_TRACE_EVENT_SIZE            CONFIG_OC_PTRACE_EVSIZE   
#endif




//
// (2) System events supported
//

// (2.1) Trace-related system events:
#ifdef CONFIG_OC_PTRACE_TREVENTS

#ifdef CONFIG_OC_PTRACE_TREVENTS_ERROR
#define _POSIX_TRACE_EVENT_ERROR
#endif

#ifdef CONFIG_OC_PTRACE_TREVENTS_START_STOP
#define _POSIX_TRACE_EVENT_START
#define _POSIX_TRACE_EVENT_STOP
#endif

#ifdef CONFIG_OC_PTRACE_TREVENTS_OVERFLOW_RESUME
#define _POSIX_TRACE_EVENT_OVERFLOW
#define _POSIX_TRACE_EVENT_RESUME
#endif

#ifdef CONFIG_OC_PTRACE_TREVENTS_EVENT_FILTER
#define _POSIX_TRACE_EVENT_SETFILTER
#endif

#else
#undef _POSIX_TRACE_EVENT_ERROR
#undef _POSIX_TRACE_EVENT_START
#undef _POSIX_TRACE_EVENT_STOP
#undef _POSIX_TRACE_EVENT_OVERFLOW
#undef _POSIX_TRACE_EVENT_RESUME
#undef _POSIX_TRACE_EVENT_SETFILTER
#endif


// (2.2) Specific system events:
#ifdef CONFIG_OC_PTRACE_SYSEVENTS
#define _POSIX_TRACE_SYSEVENTS
#endif




//
// (3) RT-Linux specific options
//

#ifdef CONFIG_OC_PTRACE_LINUXTRACE
#define _POSIX_TRACE_LINUXTRACE            200312L
#define _POSIX_TRACE_LINUX_FIFO            CONFIG_OC_PTRACE_LINUXTRACE_FIFO
#endif










//
// Nostalgic note: original POSIX definitions, no longer in use:
//
//#define _POSIX_TRACE                200009L
//#define _POSIX_TRACE_EVENT_FILTER   200009L
//#define _POSIX_TRACE_LOG            200009L
//#undef  _POSIX_TRACE_INHERITANCE    // NOT (YET) SUPPORTED!



#endif // __TRACE_POSIX_TRACE_OPTIONS_H
