/*
 * Copyright (c) 2004 Andres Terrasa.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Further, this software is distributed without any warranty that it is
 * free of the rightful claim of any third person regarding infringement
 * or the like.  Any license provided herein, whether implied or
 * otherwise, applies only to this software file.  Patent licenses, if
 * any, provided herein do not apply to combinations of this program with
 * other software, or any other product whatsoever.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Contact information: Andres Terrasa UPV
 *
 * mail: aterrasa@dsic.upv.es
 *
 */
/**********************************************************
 * 
 *    POSIX Trace TEST
 * 
 *    TEST IDENTIFIER	: ltp_ptrace_api4.c 
 * 
 *    EXECUTED BY	: root 
 * 
 *    TEST TITLE	:  POSIX Trace API 4 test
 * 
 *    TEST CASE TOTAL	: 1
 * 
 *    CPU TYPES		: i386 
 * 
 *    AUTHOR		: Andres Terrasa
 * 
 *    CO-PILOT		: Agustin Espinosa
 * 
 *    DATE STARTED	: 14/10/04
 * 
 *    INITIAL RELEASE	: RTL-PTM 1.0
 * 
 *    TEST CASES
 * 
 * 	1.-  POSIX Tracing offline retrieval API functionality.
 *	
 *    INPUT SPECIFICATIONS
 *      None
 * 
 *    OUTPUT SPECIFICATIONS
 *      None
 * 	
 *    DURATION
 * 	Terminates 
 * 
 *    SIGNALS
 *      No signals used
 *
 *    RESOURCES
 * 	RTLinux scheduler module with POSIX Tracing support enabled
 * 
 *    ENVIRONMENTAL NEEDS
 *      OCERA_DIR must be set to the ocera root directory and exported.
 *      LTPROOT must be set to the LTP root directory and exported.
 * 
 *    SPECIAL PROCEDURAL REQUIREMENTS
 * 	You need root privileges to run this test
 * 
 *    INTERCASE DEPENDENCIES
 *
 * 
 *    DETAILED DESCRIPTION
 *      This is a test for the POSIX Trace support in RT-Linux
 *
 *	This test checks the offline trace and retrieval of events.
 *
 *	In particular, the init_module creates a stream with log,
 *	creates a new (user) event type and then traces several events
 *	of this type.  Then, it creates a task which performs a flush
 *	operation on the stream.
 *
 *	The second part of the test is a Linux program called
 *	"test_ptrace_api4_analyzer.c" which opens the log and
 *	retrieves the information from it, checking that the log is
 *	correct and contains all the expected information.  Please
 *	refer to this file for more info.
 *
 *
 * 	Setup:
 * 	  Check if the test has root privileges
 * 	  Remove the RTLinux modules, if loaded.
 * 
 * 	Test:
 * 	  1.- Load the required RTLinux modules
 *        2.- Load the test module and the program to get the log
 *        3.- Run the analyzer program to check the log
 *
 * 	Cleanup:
 * 	  Remove the test module
 * 	  Print the results and exit
 *
 * 
 *#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#*#**/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <test.h>
#include <time.h>
#include <sys/klog.h>
#include <string.h>

extern int query_module(const char *name, int which,
			void *buf, size_t bufsize, size_t *ret);

/*
  LTP global variables and definitions 
*/
char *TCID = "ptrace_test_api4";/* Test program identifier. */
int TST_TOTAL = 1;		/* Total number of test cases. */
extern int Tst_count;		/* Test Case counter for tst_* routines */



/*
  Test variables and definitions
*/

#define LOG_FIFO        1
#define MODULE_DELAY    2


#define UNLOAD_RTLINUX "/sbin/rmmod rtl_sched mbuff rtl_fifo rtl_posixio \
                       rtl_time rtl >/dev/null 2>&1"

#define LOAD_TESTMOD   "/sbin/insmod \
                       $LTPROOT/testcases/bin/test_ptrace_api4_module.o \
                       delay=%d fifo=%d >/dev/null 2>&1"

#define UNLOAD_TESTMOD "/sbin/rmmod test_ptrace_api4_module >/dev/null 2>&1"

#define RUN_WLOG       "$LTPROOT/testcases/bin/wlog /dev/rtf%d \
                        test_ptrace_api4.log"

#define RUN_ANALYZER   "$LTPROOT/testcases/bin/test_ptrace_api4_analyzer \
                       test_ptrace_api4.log"



#define MSG_LOAD_FAIL   "Failed to load test_ptrace_api4_module"
#define MSG_WLOG_FAIL   "Failed to run the wlog program"
#define MSG_ANALYZER_FAIL "Failed to run the log analyzer program"
#define MSG_UNLOAD_FAIL "Failed to unload test_ptrace_api4_module"

#define MSG_TESTING1    "Starting phase 1/2: loading API 4 test module..."
#define MSG_TESTING2    "Starting phase 1/2: running log analyzer program..."
#define MSG_OK2         "Phase 2/2 passed: log successfully analyzed"
#define MSG_SUCCESS     "Offline trace/retrieval API works OK"




void cleanup ()
{
    system (UNLOAD_TESTMOD);
    system (UNLOAD_RTLINUX);
    system ("rm -f $LTPROOT/testcases/bin/test_ptrace_api4.log");
    tst_exit ();
}





void setup ()
{

    /* must run test as root */
    if (geteuid () != 0) {
	tst_brkm (TBROK, tst_exit, "You must be root to run this test");
    }

    /* If the test module is inserted, remove it first */
    tst_resm (TINFO, "Setup: unload RT-Linux and previous test modules...");
    system (UNLOAD_TESTMOD);
    system (UNLOAD_RTLINUX);

    if (query_module ("rtl", 0, NULL, 0, NULL) == 0) {
	tst_brkm (TBROK, tst_exit,
		  "RT-Linux modules already loaded: Setup failed");
    }

    /* load RTLinux modules */
    if (system ("/sbin/insmod rtl quiet=1 >/dev/null 2>&1") != 0) {
	tst_brkm (TBROK, cleanup, "Unable to load rtl module");
    }
    if (system ("/sbin/insmod rtl_time >/dev/null 2>&1") != 0) {
	tst_brkm (TBROK, cleanup, "Unable to load rtl_time module");
    }
    if (system ("/sbin/insmod rtl_posixio >/dev/null 2>&1") != 0) {
	tst_brkm (TBROK, cleanup, "Unable to load rtl_posixio module");
    }
    if (system ("/sbin/insmod rtl_fifo >/dev/null 2>&1") != 0) {
	tst_brkm (TBROK, cleanup, "Unable to load rtl_fifo module");
    }
    if (system ("/sbin/insmod mbuff >/dev/null 2>&1") != 0) {
	tst_brkm (TBROK, cleanup, "Unable to load mbuff module");
    }
    if (system ("/sbin/insmod rtl_sched >/dev/null 2>&1") != 0) {
	tst_brkm (TBROK, cleanup, "Unable to load rtl_sched module");
    }

    system("dmesg -c >/dev/null 2>&1");
}




/* load the module to be tested */
void load_module (int delay, int fifo_num)
{
    char run_wlog[512];
    char load_testmod[256];

    sprintf (load_testmod, LOAD_TESTMOD, delay, fifo_num);

    if (system (load_testmod) != 0) {
      tst_resm (TINFO, MSG_LOAD_FAIL);
    }

    sprintf(run_wlog, RUN_WLOG, fifo_num);
    if (system (run_wlog) != 0) {
      tst_brkm (TFAIL, cleanup, MSG_WLOG_FAIL);
    }

}




/* unload the module to be tested */
void unload_module ()
{
    if (system (UNLOAD_TESTMOD) != 0) {
      tst_brkm (TFAIL, cleanup, MSG_UNLOAD_FAIL);
    }
}



void run_analyzer ()
{
    if (system (RUN_ANALYZER) != 0) {
      tst_brkm (TFAIL, cleanup, MSG_ANALYZER_FAIL);
    }
}




/* capture and process the kernel module output (printk) */
char *process_module_output ()
{
    char dmesg[16 * 1024];
    char *result_line;

    klogctl (3, dmesg, 16392);
    result_line = strstr (dmesg, TCID);
    strtok (result_line, ":");
    if (strncmp (strtok (NULL, ":"), " PASS ", 6)) {
      tst_brkm (TFAIL, cleanup, strtok (NULL, "\n"));
      tst_exit ();
    }
    return strtok(NULL, "\n");
}




/* main test function */
int main (int ac, char **av, char **env)
{
    struct timespec delay, dummy;
    char *module_result_ok;

    Tst_count = 0;

    setup ();
    tst_resm (TINFO, MSG_TESTING1);
    load_module (MODULE_DELAY, LOG_FIFO);

    tst_resm (TINFO, "The module must run for 5 seconds. Waiting... ");
    delay.tv_sec = MODULE_DELAY + 5;
    delay.tv_nsec = 0;
    nanosleep (&delay, &dummy);

    module_result_ok = process_module_output();
    unload_module ();
    tst_resm (TINFO, module_result_ok);

    tst_resm (TINFO, MSG_TESTING2);
    run_analyzer ();
    tst_resm (TINFO, MSG_OK2);

    tst_resm (TPASS, MSG_SUCCESS);
    cleanup ();
    return 0;
}
