/*
   FILE:							       	
	test_ptrace_perf1_writer.c
								       	
   DESCRIPTION:							       	
        Linux analyzer of the measurements taken at the RTL side 
	(measuring trace overhead).

	It reads the measurements in shared memory (mbuff) and
	writes them to separate files (one per measurement type).
	
					       				
   Copyright (C) 2003,2004
	Andres Terrasa Barrena
	Ana Garcia Fornes
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        October 2004

*/




#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "test_ptrace_perf1.h"




FILE *open_file(char *filename) {
  FILE *f;

#ifdef _LWRITER_DEBUG
  fprintf(stderr, "linux_writer: Creating file %s\n", filename);
#endif
  if ((f=fopen(filename, "w")) == NULL) {
    fprintf(stderr, "linux_writer: Error when creating output file %s\n", 
	    filename);
    exit(1);
  }
  return f;
}




int 
main(int argc, char **argv)
{

  int mu_out = 0;
  int finish = 0;
  int type;
  long dif_time, dif_tsc;

  FILE  *trace_file, *getnext_file;
  FILE *trace1_file, *trace2_file, *getnext1_file, *getnext2_file;
  char trace1_filename[256], trace2_filename[256], getnext1_filename[256],
    getnext2_filename[256], trace_filename[256], getnext_filename[256];

  size_t struct_size;
  double mhz_fp;

  if (argc < 3) {
    fprintf(stderr, "linux_writer: Use: linux_writer output_file processor_mhz\n");
    exit(1);
  }

  mhz_fp = atof(argv[2]);


  if (mhz_fp < 0.0) {
     fprintf(stderr, "linux_writer: Error: processr_mhz %f no valid\n", 
	     mhz_fp);
     exit(1);
  }

  strncpy(trace_filename, argv[1], 256);
  strncpy(trace1_filename, argv[1], 256);
  strncpy(trace2_filename, argv[1], 256);
  strncpy(getnext_filename, argv[1], 256);
  strncpy(getnext1_filename, argv[1], 256);
  strncpy(getnext2_filename, argv[1], 256);

  strcat(trace_filename, ".tr.times");
  strcat(trace1_filename, ".tr1.times");
  strcat(trace2_filename, ".tr2.times");
  strcat(getnext_filename, ".gn.times");
  strcat(getnext1_filename, ".gn1.times");
  strcat(getnext2_filename, ".gn2.times");


  trace_file    = open_file(trace_filename);
  trace1_file   = open_file(trace1_filename);
  trace2_file   = open_file(trace2_filename);
  getnext_file  = open_file(getnext_filename);
  getnext1_file = open_file(getnext1_filename);
  getnext2_file = open_file(getnext2_filename);


  struct_size = sizeof(measure_unit_t);
  measure_shm = (volatile measure_unit_t *) mbuff_alloc(MEASUREMENTS_MBUF,
							struct_size * MAX_MEASUREMENTS);
  if( measure_shm == NULL) {
    printf("linux_writer: mbuff_alloc failed\n");
    return -1;	
  }


  do {

    type = measure_shm[mu_out].type;

    if (type != TRACE_EVENT_1 && type != GETNEXT_EVENT_1 
	&& type != TRACE_EVENT_2 && type != GETNEXT_EVENT_2 
	&& type != TRACE_SYSEVENT && type != VOID) {
#ifdef _LWRITER_DEBUG
      fprintf(stderr, 
	      "(linux_writer): Error, measurement %d in buffer with unknown type %d\n", 
	      mu_out, type);
#endif
      finish = 1;
    }
    else {
      dif_tsc = (long) (measure_shm[mu_out].after - measure_shm[mu_out].before);
      dif_time = (long) ((double) (dif_tsc * 1000) / (double) mhz_fp);

      switch (type) {
      case TRACE_EVENT_1: 
	fprintf(trace_file, "%ld\n", dif_tsc);
	fprintf(trace1_file, "%ld\n", dif_tsc);
	break;
      case TRACE_EVENT_2: 
	fprintf(trace_file, "%ld\n", dif_tsc);
	fprintf(trace2_file, "%ld\n", dif_tsc);
	break;
      case GETNEXT_EVENT_1: 
	fprintf(getnext_file, "%ld\n", dif_tsc);
	fprintf(getnext1_file, "%ld\n", dif_tsc);
	break;
      case GETNEXT_EVENT_2: 
	fprintf(getnext_file, "%ld\n", dif_tsc);
	fprintf(getnext2_file, "%ld\n", dif_tsc);
	break;
      case TRACE_SYSEVENT: 
	break;
      case VOID:           
	break;
      }
      mu_out++;
    }

  } while (mu_out < MAX_MEASUREMENTS && ! finish);



  fclose (trace_file);
  fclose (trace1_file);
  fclose (trace2_file);
  fclose (getnext_file);
  fclose (getnext1_file);
  fclose (getnext2_file);

  mbuff_free(MEASUREMENTS_MBUF,(void*)measure_shm);
  return 0;
}
