/*

   FILE:

	logto.c
								
   DESCRIPTION:								

   Copyright (C) 1998-2004
	Andres Terrasa Barrena
	Ana Garcia Fornes
	Ignacio Paches Giner
	Agustin Espinosa Minguet



  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.


  LAST REVISION:
        January 2004
*/

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h> 
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <trace.h>
#include <rtl_time.h>


//#define _LOGTO_DEBUG

#define KIWI_RELEASE       0
#define KIWI_ENDTASK       1
#define KIWI_DEADLINE      2
#define KIWI_STARTCPU      3
#define KIWI_STOPCPU       4
#define KIWI_READY         5
#define KIWI_NOTREADY      6
#define KIWI_MUTEX_LOCK    7
#define KIWI_MUTEX_UNLOCK  8
#define KIWI_EVNTUP        9
#define KIWI_EVNTDOWN      10
#define KIWI_SUSPTASK      11
#define KIWI_VLINE         12
#define KIWI_UPTEXT        13
#define KIWI_DOWNTEXT      14


#define KIWI_COLOR_CSECTION "red"

#define MAX_TASKS              50
#define MAX_MUTEX              50
#define MAX_COND               50

#define MAXSTR                 256
#define KIWI_OUTPUT            1
#define XML_OUTPUT             2
#define TEXT_OUTPUT            3

#define NONE                  -1 

#define MAGNITUDE              1


static int   task_in_csection[MAX_TASKS];
static int   linux_line = 0;
static int   kernel_line = 0;


#define LINEAR_TIME(sec,nsec)  (\
                               (hrtime_t) (((hrtime_t) sec * (hrtime_t) 1000000000)\
                                           + (hrtime_t) nsec) \
                               )


#define TIME_INITIALIZE(t)      (\
                                initial_time = \
                                   (initial_time ?  \
                                    initial_time : \
                                    (hrtime_t) t) \
                                )


#define TIME_ABS_TO_RELATIVE(t) (\
                                (hrtime_t) ((int) t ? \
                                           (t - initial_time) : 0) \
				)


#define TIME_KIWI_TIME(t)     (\
                               (long long)  ((float) (t - initial_time) \
                                           / (float) MAGNITUDE ) \
                               )

#define OUTPUT_TO_KIWI(file,t,event,task) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld  %d %d\n", \
				    (long) sec, (long) nsec, (int) event, (int) task); \
                             };

#define OUTPUT_TO_KIWI_COLOR(file,t,event,task,color) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld  %d %d \"%s\"\n", \
				    (long) sec, (long) nsec, (int) event, (int) task, (char *) color); \
                             };

#define OUTPUT_TO_KIWI2INT(file,t,event,task,arg1) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld %d %d %d\n", \
			            (long) sec, (long) nsec, (int) event, (int) task, (int) arg1); \
                             };

#define OUTPUT_TO_KIWI2STR(file,t,event,task,arg1) { \
                             long sec = (long) (t / 1000000000); \
                             long nsec = (long) (t % 1000000000); \
                             fprintf(file, "%ld.%09ld %d %d \"%s\"\n", \
			            (long) sec, (long) nsec, (int) event, (int) task, arg1); \
                             }; 






void do_calculate_maximums(trace_id_t trace_id, int verbose) {
  int                           n_events_in_log = 0;
  int                           max_tasks = 0;
  int                           max_mutexes = 0;
  int                           max_conditions = 0;
  int                           error;
  int                           end_of_log;
  struct posix_trace_event_info event;
  char                          event_data[256];
  size_t                        data_len;
  posix_trace_sysevent_data     *sysevent_data;

  //Print information to the display:
  if (verbose) {
    fprintf(stderr, "Log ok. Calculating maximum values...\n");
  }

  // First walk-through of the log: find out the max number of tasks:
  error = 0;
  end_of_log = 0;

  while (1) {
    error = posix_trace_getnext_event(trace_id, &event, 
				      event_data, sizeof(event_data), 
				      &data_len, &end_of_log);
    
    if (error || end_of_log) {
      break;
    }

    n_events_in_log++;

    if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				  POSIX_TRACE_CSWITCH) ||
	posix_trace_eventid_equal(trace_id,event.posix_event_id,
				  POSIX_TRACE_TASK_NEW_STATE) ||
	posix_trace_eventid_equal(trace_id,event.posix_event_id,
				  POSIX_TRACE_MUTEX_LOCK) ||
	posix_trace_eventid_equal(trace_id,event.posix_event_id,
				  POSIX_TRACE_MUTEX_UNLOCK) ||
	posix_trace_eventid_equal(trace_id,event.posix_event_id,
					 POSIX_TRACE_SYSCALL) ||
	posix_trace_eventid_equal(trace_id,event.posix_event_id,
					 POSIX_TRACE_COND_WAIT) ||
	posix_trace_eventid_equal(trace_id,event.posix_event_id,
				  POSIX_TRACE_COND_BROADCAST)) {

      sysevent_data = (posix_trace_sysevent_data *) event_data;
      
      if (sysevent_data->task > max_tasks) {
	max_tasks = sysevent_data->task;
      }
      
      if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				    POSIX_TRACE_MUTEX_LOCK) ||
	  posix_trace_eventid_equal(trace_id,event.posix_event_id,
				    POSIX_TRACE_MUTEX_UNLOCK)) {
	if (sysevent_data->other_data.mutex_data.mutex > max_mutexes) {
	  max_mutexes = sysevent_data->other_data.mutex_data.mutex;
	}
      }

      if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				    POSIX_TRACE_COND_WAIT) ||
	  posix_trace_eventid_equal(trace_id,event.posix_event_id,
				    POSIX_TRACE_COND_BROADCAST)) {
	if (sysevent_data->other_data.cond_data.cond > max_conditions) {
	  max_conditions = sysevent_data->other_data.cond_data.cond;
	}
      }
    }
  }

  linux_line = max_tasks + 1;

  if (verbose) {
    //Print information to the display:
    fprintf(stderr, "Results: \n");
    fprintf(stderr, "\tNumber of events in log: %d\n", n_events_in_log);
    fprintf(stderr, "\tLast registered task: %d\n", max_tasks);
    fprintf(stderr, "\tLast registered mutex: %d\n", max_mutexes);
    fprintf(stderr, "\tLast registered condition: %d\n", max_conditions);
  }



}





void do_text_output(trace_id_t trace_id, FILE *debug_file, int verbose) {
  int                           error;
  int                           end_of_log;

  struct posix_trace_status_info status;
  struct timespec               time_spec;
  char                          str[TRACE_NAME_MAX];
  size_t                        size;
  int                           pol;
  trace_attr_t                  trace_attr;

  struct posix_trace_event_info event;
  trace_event_id_t              event_id;
  char                          event_data[256];
  char                          event_name[TRACE_EVENT_NAME_MAX];
  int                           unavailable;
  size_t                        data_len;
  hrtime_t                      initial_time = (hrtime_t) 0;
  hrtime_t                      time;
  hrtime_t                      linear_time;
  int                           task_id;
  int                           mutex_id, cond_id; 
  int                           first_event = 0;
  char                          syscallstr[16];

  posix_trace_sysevent_data     *task_event;
  posix_trace_sysevent_data     *mutex_event;
  posix_trace_sysevent_data     *cond_event;
  posix_trace_sysevent_data     *syscall_event;
  posix_trace_sysevent_data     *sched_event;

  if (verbose) {
    //Print information to the display:
    fprintf(stderr, "Now generating text file... \n");
  }

  // -------------------------------------------------------------------------
  // Print attr information of the stream :
  error = posix_trace_get_attr(trace_id, &trace_attr);
  
  fprintf(debug_file, "-- [ ATTRIBUTE information ] ----------------------------------- \n");
  
  error = posix_trace_attr_getgenversion(&trace_attr, str);
  fprintf(debug_file, " Generation version     : %s\n", str);
  
  error = posix_trace_attr_getname(&trace_attr, str);
  fprintf(debug_file, " Name                   : %s\n", str);
  
  error = posix_trace_attr_getcreatetime(&trace_attr, &time_spec);
  fprintf(debug_file, " Creation time          : %ld.%ld\n",  
	  time_spec.tv_sec, time_spec.tv_nsec);
  
  error = posix_trace_attr_getclockres(&trace_attr, &time_spec);
  fprintf(debug_file, " Clock resolution       : %ld.%ld\n",  
	  time_spec.tv_sec, time_spec.tv_nsec);
  
  error = posix_trace_attr_getstreamfullpolicy(&trace_attr, &pol);
  fprintf(debug_file, " Full policy            : ");
  if (pol == POSIX_TRACE_UNTIL_FULL) {
    fprintf(debug_file, "POSIX_TRACE_UNTIL_FULL\n");
  } else if (pol == POSIX_TRACE_LOOP) {
    fprintf(debug_file, "POSIX_TRACE_LOOP\n");
  } else if (pol == POSIX_TRACE_FLUSH) {
    fprintf(debug_file, "POSIX_TRACE_FLUSH\n");
  } else {
    fprintf(debug_file, " # UNKNOWN #\n");
  }
  
  error = posix_trace_attr_getmaxusereventsize(&trace_attr, 0, &size);
  fprintf(debug_file, " Max. user event size   : %d\n", size);
  
  error = posix_trace_attr_getmaxsystemeventsize(&trace_attr, &size);
  fprintf(debug_file, " Max. system event size : %d\n", size);
  
  error = posix_trace_attr_getmaxdatasize(&trace_attr, &size);
  fprintf(debug_file, " Max data size          : %d\n", size);
  
  error = posix_trace_attr_getstreamsize(&trace_attr, &size);
  fprintf(debug_file, " Stream size            : %d\n", size);
  fprintf(debug_file, "-- [ /ATTRIBUTE information ] ---------------------------------- \n");
  
  
  // -------------------------------------------------------------------------
  // Print event types of the stream :
  fprintf(debug_file, "\n\n");
  fprintf(debug_file, "-- [ EVENT TYPES ] --------------------------------------------- \n");
  fprintf(debug_file,"EVENT TYPES:\n");
  posix_trace_eventtypelist_rewind(trace_id);
  posix_trace_eventtypelist_getnext_id(trace_id, &event_id, &unavailable);
  while (! unavailable) {
    posix_trace_eventid_get_name(trace_id,event_id,event_name);
    fprintf(debug_file," EVENT No. %3d   (%s)\n",
	    event_id, event_name);
    posix_trace_eventtypelist_getnext_id(trace_id, &event_id, &unavailable);
  } 
  fprintf(debug_file, "-- [ /EVENT TYPES ] -------------------------------------------- \n");
  
  // -------------------------------------------------------------------------
  // Print the current status of the stream :
  fprintf(debug_file, "\n\n");
  fprintf(debug_file, "-- [ STATUS information ] -------------------------------------- \n");
  error = posix_trace_get_status(trace_id, &status);
  
  fprintf(debug_file, " Stream status         : ");
  if (status.posix_stream_status == POSIX_TRACE_RUNNING) {
    fprintf(debug_file, "POSIX_TRACE_RUNNING\n");
  } else if (status.posix_stream_status == POSIX_TRACE_SUSPENDED) {
    fprintf(debug_file, "POSIX_TRACE_SUSPENDED\n");
  } else {
    fprintf(debug_file, " # UNKNOWN #\n");
  }
  
  fprintf(debug_file, " Stream full status    : ");
  if (status.posix_stream_full_status == POSIX_TRACE_FULL) {
    fprintf(debug_file, "POSIX_TRACE_FULL\n");
  } else if (status.posix_stream_full_status == POSIX_TRACE_NOT_FULL) {
    fprintf(debug_file, "POSIX_TRACE_NOT_FULL\n");
  } else {
    fprintf(debug_file, " # UNKNOWN #\n");
  }
  
  fprintf(debug_file, " Stream overrun status : ");
  if (status.posix_stream_overrun_status == POSIX_TRACE_OVERRUN) {
    fprintf(debug_file, "POSIX_TRACE_OVERRUN\n");
  } else if (status.posix_stream_overrun_status == POSIX_TRACE_NO_OVERRUN) {
    fprintf(debug_file, "POSIX_TRACE_NO_OVERRUN\n");
  } else {
    fprintf(debug_file, " # UNKNOWN #\n");
  }
  
  fprintf(debug_file, " Stream flush status   : ");
  if (status.posix_stream_flush_status == POSIX_TRACE_FLUSHING) {
    fprintf(debug_file, "POSIX_TRACE_FLUSHING\n");
  } else if (status.posix_stream_flush_status == POSIX_TRACE_NOT_FLUSHING) {
    fprintf(debug_file, "POSIX_TRACE_NOT_FLUSHING\n");
  } else {
    fprintf(debug_file, " # UNKNOWN #\n");
  }
  
  fprintf(debug_file, " Stream flush error    : %d\n",
	  status.posix_stream_flush_error);
  
  fprintf(debug_file, " Log overrun status    : ");
  if (status.posix_log_overrun_status == POSIX_TRACE_OVERRUN) {
    fprintf(debug_file, "POSIX_TRACE_OVERRUN\n");
  } else if (status.posix_log_overrun_status == POSIX_TRACE_NO_OVERRUN) {
    fprintf(debug_file, "POSIX_TRACE_NO_OVERRUN\n");
  } else {
    fprintf(debug_file, " # UNKNOWN #\n");
  }
  
  fprintf(debug_file, " Log full status       : ");
  if (status.posix_log_overrun_status == POSIX_TRACE_FULL) {
    fprintf(debug_file, "POSIX_TRACE_FULL\n");
  } else if (status.posix_log_overrun_status == POSIX_TRACE_NOT_FULL) {
    fprintf(debug_file, "POSIX_TRACE_NOT_FULL\n");
  } else {
    fprintf(debug_file, " # UNKNOWN #\n");
  }
  fprintf(debug_file, "-- [ /STATUS information ] ------------------------------------- \n");
  
  
  fprintf(debug_file, "\n\n");
  fprintf(debug_file, "-- [ BEGIN TRACING ] ------------------------------------------- \n");
  


  posix_trace_rewind(trace_id);
  error = 0;
  end_of_log = 0;

  while (1) {
    error = posix_trace_getnext_event(trace_id, &event, 
				      event_data, sizeof(event_data), 
				      &data_len, &end_of_log);
    

    if (error || end_of_log) {
      break;
    }

    linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
			      event.posix_timestamp.tv_nsec);
    
    if (!first_event) {
      TIME_INITIALIZE(linear_time);
      first_event = 1;
    }      
    
    posix_trace_eventid_get_name(trace_id,event.posix_event_id,event_name);
    
    // Now switch depending on the type of event:
    
    // POSIX_TRACE_START
    if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				  POSIX_TRACE_START)) {
      fprintf(debug_file, 
	      "START         (%ld.%09ld)\n", 
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    }
    
    // POSIX_TRACE_STOP
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_STOP)) {
      fprintf(debug_file, 
	      "STOP          (%ld.%09ld)\n",
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    }
    
    // POSIX_TRACE_ERROR
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_ERROR)) {
      fprintf(debug_file, 
	      "ERROR         (%ld.%09ld)\n",
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    } 
    
    // POSIX_TRACE_OVERFLOW
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_OVERFLOW)) {
      fprintf(debug_file, 
	      "OVERFLOW      (%ld.%09ld)\n",
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    } 
    
    // POSIX_TRACE_RESUME
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_RESUME)) {
      fprintf(debug_file, 
	      "RESUME        (%ld.%09ld)\n",
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    } 
    
    // POSIX_TRACE_FILTER
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_FILTER)) {
      fprintf(debug_file, 
	      "FILTER        (%ld.%09ld)\n",
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    }
    
    // POSIX_TRACE_CSWITCH
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_CSWITCH)) {
      
      task_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id = (task_event->task == 0?
		 linux_line : 
		 task_event->task);
      
      if (task_id != NONE) {
	
	  fprintf(debug_file,
		  "TASK_START    (%ld.%09ld):  task=%d  t=%llu\n", 
		  event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,  
		  task_id, time); 
      }
    }
  
    // POSIX_TRACE_TASK_NEW_STATE
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_TASK_NEW_STATE)) {
      
      task_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      task_id     = task_event->task;
      
      if (task_id != NONE) {
	
	switch(task_event->index) {
	  
	case TASK_NEW_STATE_RELEASED : 	  
	  fprintf(debug_file, 
		  "TASK_STATE    (%ld.%09ld):  task=%d  state=RELEASE\n",
		  event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		  task_id);
	  break;
	  
	case TASK_NEW_STATE_READY :  
	  fprintf(debug_file, 
		  "TASK_STATE    (%ld.%09ld):  task=%d  state=READY\n",
		  event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		  task_id);
	  break;
	  
	case TASK_NEW_STATE_UNREADY :  
	  fprintf(debug_file, 
		  "TASK_STATE    (%ld.%09ld):  task=%d  state=SUSPENDED\n",
		  event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		  task_id);
	  break;
	}
      }
    }
    
    
    // POSIX_TRACE_MUTEX_LOCK
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_MUTEX_LOCK)) {
      
      mutex_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = mutex_event->task;
      mutex_id   = mutex_event->other_data.mutex_data.mutex;
	      
      if (task_id != NONE && mutex_id != NONE) {
	
	switch(mutex_event->index) {
	  
	case MUTEX_LOCK_DENIED: 
	  fprintf(debug_file,
		  "MUTEX_LOCK    (%ld.%09ld):  task=%d  mutex=%d  t=%llu  (LOCK DENIED)\n", 
		  event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,	  
		  task_id, mutex_id, time); 
	  break;
	  
	case MUTEX_LOCK_GRANTED :
	  fprintf(debug_file,
		  "MUTEX_LOCK    (%ld.%09ld):  task=%d  mutex=%d  t=%llu  (LOCK GRANTED)\n", 
		  event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		  task_id, mutex_id, time); 
	  break;
	}
      }
    }
    
    
    // POSIX_TRACE_MUTEX_UNLOCK
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_MUTEX_UNLOCK)) {
      
      mutex_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = mutex_event->task;
      mutex_id   =  mutex_event->other_data.mutex_data.mutex;
      
      if (task_id != NONE && mutex_id != NONE) {
	fprintf(debug_file,
		"MUTEX_UNLOCK  (%ld.%09ld):  task=%d  mutex=%d  t=%llu\n", 
		event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		task_id, mutex_id, time); 
      }
    }
    
    
    // POSIX_TRACE_SYSCALL
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_SYSCALL)) {
      
      syscall_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id = (syscall_event->task == 0?
		 linux_line : 
		 syscall_event->task);
      
      if (task_id != NONE) {
	
	switch(syscall_event->index) {
	  
	case SYSCALL_PTHREAD_CREATE :
	  sprintf(syscallstr,"CRET");
	  break;
	  
	case SYSCALL_SCHED_YIELD :
	  sprintf(syscallstr,"YILD");
	  break;
	  
	case SYSCALL_PTHREAD_KILL: 
	  sprintf(syscallstr,"KILL");
	  break;
	  
	case SYSCALL_PTHREAD_EXIT:
	  sprintf(syscallstr,"EXT");
	  break;
	  
	case SYSCALL_PTHREAD_WAIT_NP:
	  sprintf(syscallstr,"WAIT");
	  break;
	  
	case SYSCALL_PTHREAD_SUSPEND_NP:
	  sprintf(syscallstr,"SUSP");
	  break;
	  
	case SYSCALL_PTHREAD_WAKEUP_NP:
	  sprintf(syscallstr,"WKUP");
	  break;
	  
	case SYSCALL_PTHREAD_DELETE_NP:
	  sprintf(syscallstr,"DLET");
	  break;
	  
	case SYSCALL_CLOCK_NANOSLEEP:
	  sprintf(syscallstr,"SLEP");
	  break;
	  
	case SYSCALL_PTHREAD_MUTEX_LOCK:
	  sprintf(syscallstr,"LOCK");
	  break;
	  
	case SYSCALL_PTHREAD_MUTEX_UNLOCK:
	  sprintf(syscallstr,"ULCK");
	  break;
	  
	case SYSCALL_PTHREAD_COND_WAIT:
	  sprintf(syscallstr,"WAIT");
	  break;
	  
	case SYSCALL_PTHREAD_COND_SIGNAL:
	  sprintf(syscallstr,"SIGL");
	  break;
	  
	case SYSCALL_PTHREAD_COND_BROADCAST:
	  sprintf(syscallstr,"BCST");
	  break;
	  //default:
	}
	
	fprintf(debug_file,
		"SYSTEM_CALL   (%ld.%09ld):  syscall=%s  task=%d  t=%llu   (addr=%x)\n", 
		event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		syscallstr, task_id, time, (int) event.posix_prog_address); 
      }
    }
    

    // POSIX_TRACE_SYSCALL_END
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_SYSCALL_END)) {
      
      syscall_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id = (syscall_event->task == 0?
		 linux_line : 
		 syscall_event->task);
      
      if (task_id != NONE) {
	
	switch(syscall_event->index) {
	  
	case SYSCALL_PTHREAD_CREATE :
	  sprintf(syscallstr,"CRET");
	  break;
	  
	case SYSCALL_SCHED_YIELD :
	  sprintf(syscallstr,"YILD");
	  break;
	  
	case SYSCALL_PTHREAD_KILL: 
	  sprintf(syscallstr,"KILL");
	  break;
	  
	case SYSCALL_PTHREAD_EXIT:
	  sprintf(syscallstr,"EXT");
	  break;
	  
	case SYSCALL_PTHREAD_WAIT_NP:
	  sprintf(syscallstr,"WAIT");
	  break;
	  
	case SYSCALL_PTHREAD_SUSPEND_NP:
	  sprintf(syscallstr,"SUSP");
	  break;
	  
	case SYSCALL_PTHREAD_WAKEUP_NP:
	  sprintf(syscallstr,"WKUP");
	  break;
	  
	case SYSCALL_PTHREAD_DELETE_NP:
	  sprintf(syscallstr,"DLET");
	  break;
	  
	case SYSCALL_CLOCK_NANOSLEEP:
	  sprintf(syscallstr,"SLEP");
	  break;
	  
	case SYSCALL_PTHREAD_MUTEX_LOCK:
	  sprintf(syscallstr,"LOCK");
	  break;
	  
	case SYSCALL_PTHREAD_MUTEX_UNLOCK:
	  sprintf(syscallstr,"ULCK");
	  break;
	  
	case SYSCALL_PTHREAD_COND_WAIT:
	  sprintf(syscallstr,"WAIT");
	  break;
	  
	case SYSCALL_PTHREAD_COND_SIGNAL:
	  sprintf(syscallstr,"SIGL");
	  break;
	  
	case SYSCALL_PTHREAD_COND_BROADCAST:
	  sprintf(syscallstr,"BCST");
	  break;
	  //default:
	}
	
	fprintf(debug_file,
		"SYS_CALL_END  (%ld.%09ld):  syscall=%s  task=%d  t=%llu   (addr=%x)\n", 
		event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		syscallstr, task_id, time, (int) event.posix_prog_address); 
      }
    }
    

    // POSIX_TRACE_SCHED_START
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_SCHED_START)) {
      
      sched_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      fprintf(debug_file,
	      "SCHED_START   (%ld.%09ld):  \n", 
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    }
    
    // POSIX_TRACE_SCHED_END
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_SCHED_END)) {
      
      sched_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      fprintf(debug_file,
	      "SCHED_END     (%ld.%09ld):  \n", 
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    }
    
    // POSIX_TRACE_COND_WAIT
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_COND_WAIT)) {
      
      cond_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = cond_event->task;
      mutex_id   = cond_event->other_data.cond_data.mutex;
      cond_id    = cond_event->other_data.cond_data.cond;
      
      if (task_id != NONE && mutex_id != NONE && cond_id != NONE) {
	fprintf(debug_file,
		"COND_WAIT  (%ld.%09ld):  task=%d  cond=%d  mutex=%d  t=%llu\n", 
		event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
		task_id, cond_id, mutex_id, time);
      
      }
    }
    
    
    // POSIX_TRACE_COND_BROADCAST
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_COND_BROADCAST)) {
      
      cond_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = cond_event->task;
      mutex_id   = cond_event->other_data.cond_data.mutex;
      cond_id    = cond_event->other_data.cond_data.cond;
      if (task_id != NONE && mutex_id != NONE && cond_id != NONE) {
	fprintf(debug_file,
		"COND_BROAD (%ld.%09ld): task=%d  cond=%d  mutex=%d  t=%llu\n", 
		event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,  
		task_id, cond_id, mutex_id, time);
      }
    } else {
      
      fprintf(debug_file,
	      "OTHER_EVENT  (%ld.%09ld): %d\n", 
	      event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec,
	      event.posix_event_id);
      
    }
    
  } // while

  fprintf(debug_file, "-- [ END  TRACING ] ------------------------------------------- \n");


  if (verbose) {
    //Print information to the display:
    fprintf(stderr, "Done.\n");
  }
}































void do_kiwi_output(trace_id_t trace_id, FILE *kiwi_file, int verbose) {
  int                           error;
  int                           end_of_log;

  struct posix_trace_event_info event;
  char                          event_data[256];
  char                          event_name[TRACE_EVENT_NAME_MAX];
  size_t                        data_len;
  hrtime_t                      initial_time = (hrtime_t) 0;
  hrtime_t                      last_overhead_start;
  hrtime_t                      time;
  hrtime_t                      linear_time;
  int                           task_id;
  int                           mutex_id, cond_id; 
  int                           i;
  int                           current_task = NONE;
  int                           first_event = 0;
  char                          syscallstr[16];
  char                          objstr[16];
  posix_trace_sysevent_data     *task_event;
  posix_trace_sysevent_data     *mutex_event;
  posix_trace_sysevent_data     *cond_event;
  posix_trace_sysevent_data     *syscall_event;
  posix_trace_sysevent_data     *sched_event;


  for (i=0; i < MAX_TASKS; i++)
    task_in_csection[i] = 0;
  
  fprintf(kiwi_file, "LINE_NAME %d Kernel\n", kernel_line);
  fprintf(kiwi_file, "LINE_NAME %d Linux\n", linux_line);

  if (verbose) {
    //Print information to the display:
    fprintf(stderr, "Now generating kiwi file... \n");
  }

  // Rewind the log file in order to process the events:
  posix_trace_rewind(trace_id);
  error = 0;
  end_of_log = 0;

  while (1) {
    error = posix_trace_getnext_event(trace_id, &event, 
				      event_data, sizeof(event_data), 
				      &data_len, &end_of_log);
    

    if (error || end_of_log) {
      break;
    }

    linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
			      event.posix_timestamp.tv_nsec);
    
    if (!first_event) {
      TIME_INITIALIZE(linear_time);
      first_event = 1;
    }      
    
    posix_trace_eventid_get_name(trace_id,event.posix_event_id,event_name);
    
    // Now switch depending on the type of event:
    
    // POSIX_TRACE_START
    if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				  POSIX_TRACE_START)) {
    }
    
    // POSIX_TRACE_STOP
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_STOP)) {
    }
    
    // POSIX_TRACE_ERROR
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_ERROR)) {
    } 
    
    // POSIX_TRACE_OVERFLOW
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_OVERFLOW)) {
    } 
    
    // POSIX_TRACE_RESUME
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_RESUME)) {
    } 
    
    // POSIX_TRACE_FILTER
    else if (posix_trace_eventid_equal(trace_id, event.posix_event_id,
				       POSIX_TRACE_FILTER)) {
    }
    
    // POSIX_TRACE_CSWITCH
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_CSWITCH)) {
      
      task_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id = (task_event->task == 0?
		 linux_line : 
		 task_event->task);
      
      if (task_id != NONE) {
	
	if (current_task != NONE) {
	  if (task_in_csection[current_task]) {
	    OUTPUT_TO_KIWI_COLOR(kiwi_file,time,KIWI_STOPCPU, current_task, 
				 KIWI_COLOR_CSECTION);
	  } else {
	    OUTPUT_TO_KIWI(kiwi_file,time,KIWI_STOPCPU, current_task);
	  }
	}
	current_task = task_id;
	
	OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STARTCPU, task_id);
	OUTPUT_TO_KIWI2INT(kiwi_file, time, KIWI_UPTEXT, task_id,
			   task_event->priority);
      }
    }
    
    // POSIX_TRACE_TASK_NEW_STATE
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_TASK_NEW_STATE)) {
      
      task_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      task_id     = task_event->task;
      
      if (task_id != NONE) {
	
	switch(task_event->index) {
	  
	case TASK_NEW_STATE_RELEASED : 	  
	  OUTPUT_TO_KIWI(kiwi_file, time, KIWI_RELEASE, task_id);
	  //OUTPUT_TO_KIWI(kiwi_file, time, KIWI_READY, task_id);
	  break;
	case TASK_NEW_STATE_READY :  
	  OUTPUT_TO_KIWI(kiwi_file, time, KIWI_READY, task_id);
	  break;
	  
	case TASK_NEW_STATE_UNREADY :  
	  OUTPUT_TO_KIWI(kiwi_file, time, KIWI_NOTREADY, task_id);
	  break;
	}
      }
    }
    
    
    // POSIX_TRACE_MUTEX_LOCK
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_MUTEX_LOCK)) {
      
      mutex_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = mutex_event->task;
      mutex_id   = mutex_event->other_data.mutex_data.mutex;
	      
      if (task_id != NONE && mutex_id != NONE) {
	
	switch(mutex_event->index) {
	  
	case MUTEX_LOCK_DENIED: 
	  OUTPUT_TO_KIWI(kiwi_file, time, KIWI_SUSPTASK, task_id);
	  break;
	  
	case MUTEX_LOCK_GRANTED :
	  OUTPUT_TO_KIWI(kiwi_file, time, KIWI_MUTEX_LOCK, task_id);
	  sprintf(objstr, "  m%d", mutex_id); 
	  OUTPUT_TO_KIWI2STR(kiwi_file, time, KIWI_UPTEXT, task_id,
			     objstr);
	  OUTPUT_TO_KIWI(kiwi_file,time,KIWI_STOPCPU, task_id);
	  OUTPUT_TO_KIWI(kiwi_file,time,KIWI_STARTCPU, task_id);
	  
	  task_in_csection[task_id] = 1;
	  break;
	}
      }
    }
    
    
    // POSIX_TRACE_MUTEX_UNLOCK
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_MUTEX_UNLOCK)) {
      
      mutex_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = mutex_event->task;
      mutex_id   =  mutex_event->other_data.mutex_data.mutex;
      
      if (task_id != NONE && mutex_id != NONE) {
	OUTPUT_TO_KIWI(kiwi_file, time, KIWI_MUTEX_UNLOCK, task_id);
	sprintf(objstr, "  m%d", mutex_id); 
	OUTPUT_TO_KIWI2STR(kiwi_file, time, KIWI_UPTEXT, task_id,
			   objstr);
	OUTPUT_TO_KIWI_COLOR(kiwi_file,time,KIWI_STOPCPU, task_id, KIWI_COLOR_CSECTION);
	OUTPUT_TO_KIWI(kiwi_file,time,KIWI_STARTCPU, task_id);
	
	task_in_csection[task_id] = 0;	
      }
    }
    
    
    
    // POSIX_TRACE_SYSCALL
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_SYSCALL)) {
      
      syscall_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id = (syscall_event->task == 0?
		 linux_line : 
		 syscall_event->task);
      
      if (task_id != NONE) {
	
	switch(syscall_event->index) {
	  
	case SYSCALL_PTHREAD_CREATE :
	  sprintf(syscallstr,"CRET");
	  break;
	  
	case SYSCALL_SCHED_YIELD :
	  sprintf(syscallstr,"YILD");
	  break;
	  
	case SYSCALL_PTHREAD_KILL: 
	  sprintf(syscallstr,"KILL");
	  break;
	  
	case SYSCALL_PTHREAD_EXIT:
	  sprintf(syscallstr,"EXT");
	  break;
	  
	case SYSCALL_PTHREAD_WAIT_NP:
	  sprintf(syscallstr,"WAIT");
	  break;
	  
	case SYSCALL_PTHREAD_SUSPEND_NP:
	  sprintf(syscallstr,"SUSP");
	  break;
	  
	case SYSCALL_PTHREAD_WAKEUP_NP:
	  sprintf(syscallstr,"WKUP");
	  break;
	  
	case SYSCALL_PTHREAD_DELETE_NP:
	  sprintf(syscallstr,"DLET");
	  break;
	  
	case SYSCALL_CLOCK_NANOSLEEP:
	  sprintf(syscallstr,"SLEP");
	  break;
	  
	case SYSCALL_PTHREAD_MUTEX_LOCK:
	  sprintf(syscallstr,"LOCK");
	  break;
	  
	case SYSCALL_PTHREAD_MUTEX_UNLOCK:
	  sprintf(syscallstr,"ULCK");
	  break;
	  
	case SYSCALL_PTHREAD_COND_WAIT:
	  sprintf(syscallstr,"WAIT");
	  break;
	  
	case SYSCALL_PTHREAD_COND_SIGNAL:
	  sprintf(syscallstr,"SIGL");
	  break;
	  
	case SYSCALL_PTHREAD_COND_BROADCAST:
	  sprintf(syscallstr,"BCST");
	  break;
	  //default:
	}
	
	OUTPUT_TO_KIWI2STR(kiwi_file, time, KIWI_EVNTDOWN, task_id,
			   syscallstr);
      }
    }
    
    // POSIX_TRACE_SCHED_START
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_SCHED_START)) {
      
      sched_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STARTCPU,  kernel_line);
      if (sched_event->index == 1) {
	OUTPUT_TO_KIWI2STR(kiwi_file, time, KIWI_EVNTUP, kernel_line, "INT");
      }
      last_overhead_start = time;
    }	    
    
    // POSIX_TRACE_SCHED_END
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_SCHED_END)) {
      
      sched_event = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      OUTPUT_TO_KIWI(kiwi_file, time, KIWI_STOPCPU,   kernel_line);      
    }
    
    // POSIX_TRACE_COND_WAIT
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_COND_WAIT)) {
      
      cond_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = cond_event->task;
      mutex_id   = cond_event->other_data.cond_data.mutex;
      cond_id    = cond_event->other_data.cond_data.cond;
      
      if (task_id != NONE && mutex_id != NONE && cond_id != NONE) {
      }
    }
    
    
    // POSIX_TRACE_COND_BROADCAST
    else if (posix_trace_eventid_equal(trace_id,event.posix_event_id,
				       POSIX_TRACE_COND_BROADCAST)) {
      
      cond_event  = (posix_trace_sysevent_data *) event_data;
      
      linear_time = LINEAR_TIME(event.posix_timestamp.tv_sec,
				event.posix_timestamp.tv_nsec);
      
      time        = TIME_KIWI_TIME(linear_time);
      
      task_id    = cond_event->task;
      mutex_id   = cond_event->other_data.cond_data.mutex;
      cond_id    = cond_event->other_data.cond_data.cond;
      if (task_id != NONE && mutex_id != NONE && cond_id != NONE) {
      }
    }
    
  } // while
  
  if (verbose) {
    //Print information to the display:
    fprintf(stderr, "Done.\n");
  }
  
  if (current_task != NONE) {
    OUTPUT_TO_KIWI(kiwi_file,time,KIWI_STOPCPU, current_task);
  }
}




















void do_xml_output(trace_id_t trace_id, FILE *xml_file, int verbose) {
  int                           error;
  int                           end_of_log;

  struct posix_trace_status_info status;
  struct timespec               time_spec;
  char                          str[TRACE_NAME_MAX];
  char                          polstr[TRACE_NAME_MAX];
  size_t                        size;
  int                           pol;
  trace_attr_t                  trace_attr;

  struct posix_trace_event_info event;
  trace_event_id_t              event_id;
  char                          event_data[256];
  char                          event_name[TRACE_EVENT_NAME_MAX];
  int                           unavailable;
  size_t                        data_len;

  /*
  int                           task_id;
  int                           mutex_id;
  int                           cond_id; 
  char                          syscallstr[16];
  posix_trace_sysevent_data     *syscall_event;
  */

  if (verbose) {
    //Print information to the display:
    fprintf(stderr, "Now generating xml file... \n");
  }

  error = posix_trace_get_attr(trace_id, &trace_attr);
  
  fprintf(xml_file, "<posix_trace_log>\n");


  // BEGIN: Log header
  fprintf(xml_file, "  <log_header>\n");


  // BEGIN: Stream attributes
  fprintf(xml_file, "    <stream_attribute>\n");

  error = posix_trace_attr_getgenversion(&trace_attr, str);
  fprintf(xml_file, "      <generation-version>%s</generation-version>\n", 
	  str);
  
  error = posix_trace_attr_getname(&trace_attr, str);
  fprintf(xml_file, "      <trace-name>%s</trace-name>\n", str);

  error = posix_trace_attr_getcreatetime(&trace_attr, &time_spec);
  fprintf(xml_file, "      <creation-time>%ld.%09ld</creation-time>\n", 
	  time_spec.tv_sec, time_spec.tv_nsec);

  error = posix_trace_attr_getclockres(&trace_attr, &time_spec);
  fprintf(xml_file, "      <clock-resolution>%ld.%09ld</clock-resolution>\n", 
	  time_spec.tv_sec, time_spec.tv_nsec);

  error = posix_trace_attr_getstreamsize(&trace_attr, &size);
  fprintf(xml_file, "      <stream-min-size>%d</stream-min-size>\n", 
	  (int) size);

  error = posix_trace_attr_getstreamfullpolicy(&trace_attr, &pol);
  switch (pol) {
  case POSIX_TRACE_LOOP:
    sprintf(polstr, "POSIX_TRACE_LOOP");
    break;
  case POSIX_TRACE_UNTIL_FULL:
    sprintf(polstr, "POSIX_TRACE_UNTIL_FULL");
    break;
  case POSIX_TRACE_FLUSH:
    sprintf(polstr, "POSIX_TRACE_FLUSH");
    break;
  default:
    sprintf(polstr, "#UNKNOWN#");
  }
  fprintf(xml_file, "      <stream-full-policy>%s</stream-full-policy>\n", 
	  polstr);

  error = posix_trace_attr_getmaxdatasize(&trace_attr, &size);
  fprintf(xml_file, "      <max-data-size>%d</max-data-size>\n", (int) size);
  /*
  error = posix_trace_attr_getinherited(&trace_attr, &pol);
  switch (pol) {
  case POSIX_TRACE_CLOSE_FOR_CHILD:
    sprintf(polstr, "POSIX_TRACE_CLOSE_FOR_CHILD");
    break;
  case POSIX_TRACE_UNTIL_INHERITED:
    sprintf(polstr, "POSIX_TRACE_INHERITED");
    break;
  default:
    sprintf(polstr, "#UNKNOWN#");
  }
  fprintf(xml_file, "      <inherited>%s</inherited>\n", polstr);
  */
  /*
  error = posix_trace_attr_getlogsize(&trace_attr, &size);
  fprintf(xml_file, "      <log-max-size>%d</log-max-size>\n", (int) size);

  error = posix_trace_attr_getlogfullpolicy(&trace_attr, &pol);
  switch (pol) {
  case POSIX_TRACE_LOOP:
    sprintf(polstr, "POSIX_TRACE_LOOP");
    break;
  case POSIX_TRACE_UNTIL_FULL:
    sprintf(polstr, "POSIX_TRACE_UNTIL_FULL");
    break;
  case POSIX_TRACE_APPEND:
    sprintf(polstr, "POSIX_TRACE_APPEND");
    break;
  default:
    sprintf(polstr, "#UNKNOWN#");
  }
  fprintf(xml_file, "      <log-full-policy>%s</log-full-policy>\n", polstr);
  */
  fprintf(xml_file, "    </stream_attribute>\n");
  // END: Stream attributes


  // BEGIN: Stream event types
  fprintf(xml_file, "    <event_types_list>\n");

  posix_trace_eventtypelist_rewind(trace_id);
  posix_trace_eventtypelist_getnext_id(trace_id, &event_id, &unavailable);
  while (! unavailable) {
    posix_trace_eventid_get_name(trace_id,event_id,event_name);
    fprintf(xml_file,"      <event_type>\n");
    fprintf(xml_file,"        <posix_event_id>%d</posix_event_id>\n", event_id);
    fprintf(xml_file,"        <posix_event_name>%s</posix_event_name>\n", 
	    event_name);
    fprintf(xml_file,"      </event_type>\n");
    posix_trace_eventtypelist_getnext_id(trace_id, &event_id, &unavailable);
  } 
  fprintf(xml_file, "    </event_types_list>\n");
  // END: Stream event types

  
  // BEGIN: Stream status
  error = posix_trace_get_status(trace_id, &status);
  fprintf(xml_file, "    <stream_status>\n");

  if (status.posix_stream_status == POSIX_TRACE_RUNNING) {
    sprintf(polstr, "POSIX_TRACE_RUNNING");
  } else if (status.posix_stream_status == POSIX_TRACE_SUSPENDED) {
    sprintf(polstr, "POSIX_TRACE_SUSPENDED");
  } else {
    sprintf(polstr, " # UNKNOWN #");
  }
  fprintf(xml_file, "      <posix-stream-status>%s</posix-stream-status>\n",
	  polstr);


  if (status.posix_stream_full_status == POSIX_TRACE_FULL) {
    sprintf(polstr, "POSIX_TRACE_FULL");
  } else if (status.posix_stream_full_status == POSIX_TRACE_NOT_FULL) {
    sprintf(polstr, "POSIX_TRACE_NOT_FULL");
  } else {
    sprintf(polstr, " # UNKNOWN #\n");
  }
  fprintf(xml_file, "      <posix-stream-full-status>%s</posix-stream-full-status>\n",
	  polstr);
  
  if (status.posix_stream_overrun_status == POSIX_TRACE_OVERRUN) {
    sprintf(polstr, "POSIX_TRACE_OVERRUN");
  } else if (status.posix_stream_overrun_status == POSIX_TRACE_NO_OVERRUN) {
    sprintf(polstr, "POSIX_TRACE_NO_OVERRUN");
  } else {
    sprintf(polstr, " # UNKNOWN #");
  }
  fprintf(xml_file, "      <posix-stream-overrun-status>%s</posix-stream-overrun-status>\n",
	  polstr);
 
  if (status.posix_stream_flush_status == POSIX_TRACE_FLUSHING) {
    sprintf(polstr, "POSIX_TRACE_FLUSHING");
  } else if (status.posix_stream_flush_status == POSIX_TRACE_NOT_FLUSHING) {
    sprintf(polstr, "POSIX_TRACE_NOT_FLUSHING");
  } else {
    sprintf(polstr, " # UNKNOWN #");
  }
  fprintf(xml_file, "      <posix-stream-flush-status>%s</posix-stream-flush-status>\n",
	  polstr);
  
  fprintf(xml_file, "      <posix-stream-flush-error>%d</posix-stream-flush-error>\n",
	  status.posix_stream_flush_error);
  
  if (status.posix_log_overrun_status == POSIX_TRACE_OVERRUN) {
    sprintf(polstr, "POSIX_TRACE_OVERRUN");
  } else if (status.posix_log_overrun_status == POSIX_TRACE_NO_OVERRUN) {
    sprintf(polstr, "POSIX_TRACE_NO_OVERRUN");
  } else {
    sprintf(polstr, " # UNKNOWN #");
  }
  fprintf(xml_file, "      <posix-log-overrun-status>%s</posix-log-overrun-status>\n",
	  polstr);
  
  if (status.posix_log_overrun_status == POSIX_TRACE_FULL) {
    sprintf(polstr, "POSIX_TRACE_FULL");
  } else if (status.posix_log_overrun_status == POSIX_TRACE_NOT_FULL) {
    sprintf(polstr, "POSIX_TRACE_NOT_FULL");
  } else {
    sprintf(polstr, " # UNKNOWN #");
  }
  fprintf(xml_file, "      <posix-log-full-status>%s</posix-log-full-status>\n",
	  polstr);

  fprintf(xml_file, "    </stream_status>\n");
  // END: Stream status


  fprintf(xml_file, "  </log_header>\n");
  // END: Log header



  // BEGIN: Events
  fprintf(xml_file, "  <log_events>\n");


  // Rewind the log file in order to process the events:
  posix_trace_rewind(trace_id);
  error = 0;
  end_of_log = 0;

  while (1) {
    error = posix_trace_getnext_event(trace_id, &event, 
				      event_data, sizeof(event_data), 
				      &data_len, &end_of_log);
    

    if (error || end_of_log) {
      break;
    }

    fprintf(xml_file, "    <trace_event>\n");
    
    fprintf(xml_file, "      <posix-event-id>%d</posix-event-id>\n", 
	    event.posix_event_id);
    fprintf(xml_file, "      <posix-pid>%d</posix-pid>\n", 
	    event.posix_pid);
    fprintf(xml_file, "      <posix-prog-address>%x</posix-prog-address>\n", 
	    (int) event.posix_prog_address);
    fprintf(xml_file, "      <posix-truncation-status>%x</posix-truncation-status>\n", 
	    event.posix_truncation_status);
    fprintf(xml_file, "      <posix-timestamp>%ld.%09ld</posix-timestamp>\n", 
	    event.posix_timestamp.tv_sec, event.posix_timestamp.tv_nsec);
    fprintf(xml_file, "      <posix-thread-id>%d</posix-thread-id>\n", 
	    (int) event.posix_thread_id);

    
    fprintf(xml_file, "      <event-data>\n");

    switch (event.posix_event_id) {

    case  POSIX_TRACE_START :
      break;

    case  POSIX_TRACE_STOP :
      fprintf(xml_file, "        <auto>%d</auto>\n", 
	      (int) *((int *) event_data));
      break;

    case  POSIX_TRACE_OVERFLOW :
      break;

    case  POSIX_TRACE_RESUME :
      break;

    case  POSIX_TRACE_FILTER :
      break;

    case  POSIX_TRACE_CSWITCH :
      break;

    case  POSIX_TRACE_TASK_NEW_STATE :
      break;

    case  POSIX_TRACE_SYSCALL_END :
      break;

    case  POSIX_TRACE_SCHED_START :
      break;

    case  POSIX_TRACE_SCHED_END :
      break;

    case  POSIX_TRACE_MUTEX_LOCK :
      break;

    case  POSIX_TRACE_MUTEX_UNLOCK :
      break;

    case  POSIX_TRACE_COND_WAIT :
      break;

    case  POSIX_TRACE_COND_BROADCAST :
      break;


    }


    fprintf(xml_file, "      </event-data>\n");

    fprintf(xml_file, "    </trace_event>\n");

  }



  fprintf(xml_file, "  </log_events>\n");
  // END: Events


  fprintf(xml_file, "</posix_trace_log>\n");
  // END: Log 

  

  
}








void do_print_help (void) {

  fprintf(stderr,"usage: logto FORMAT [OPTION...] logfile\n"  );
  fprintf(stderr,"FORMATs available: \n");
  fprintf(stderr,"      kiwi, xml, text\n");
  fprintf(stderr,"OPTIONs available: \n");
  fprintf(stderr,"  -o  outputfile    file to write the output\n");
  fprintf(stderr,"                    (otherwise, standard output is used)\n");
  fprintf(stderr,"  -h                print this help and quit\n");
  fprintf(stderr,"  -v                be verbose about the processing\n");
  fprintf(stderr,"                    (on the standard error output)\n");

}













int main(int argc, char *argv[]) {

  char                          log_filename[MAXSTR];
  int                           log_filedesc;

  char                          output_filename[MAXSTR];

  FILE                          *output_file;

  int                           output  = 0;
  int                           verbose = 0;
  int                           error   = 0;
  int                           arg_i;

  trace_id_t                    trace_id;




  
  // Command-line argument processing:
  if (argc < 3) {
    do_print_help();
    exit (1);
  } else {
    // First argument:
    if (! strcmp(argv[1], "kiwi")) {
      output = KIWI_OUTPUT;
    } else if  (! strcmp(argv[1], "xml")) {
      output = XML_OUTPUT;
    } else if  (! strcmp(argv[1], "text")) {
      output = TEXT_OUTPUT;
    } else {
      do_print_help();
      exit (1);
    }

    strncpy(output_filename,"stdout",MAXSTR);
    // Second argument:
    if (argc == 3) {
      strncpy(log_filename,argv[2],MAXSTR);
    } else {
      // More arguments (let's see how many):
      arg_i = 2;
      while (arg_i <= argc - 2) {
	if (! strcmp(argv[arg_i], "-o")) {
	  arg_i++;
	  strncpy(output_filename,argv[arg_i],MAXSTR);
	  arg_i++;
	} else if (! strcmp(argv[arg_i], "-v")) {
	  arg_i++;
	  verbose = 1;
	} else if (! strcmp(argv[arg_i], "-h")) {
	  arg_i++;
	  do_print_help();
	} else {
	  do_print_help();
	  exit (1);
	}
      }
      if (arg_i == argc - 1) {
	strncpy(log_filename,argv[arg_i],MAXSTR);
      } else {
	do_print_help();
	exit (1);
      }
    }
  }

#ifdef _LOGTO_DEBUG
  fprintf(stderr,"output = %d\n", output);
  fprintf(stderr,"output_filename = %s\n", output_filename);
  fprintf(stderr,"verbose = %d\n", verbose);
  fprintf(stderr, "logfile = %s\n", log_filename);
#endif


  // Creat/open the output file:
  if (!strcmp (output_filename, "stdout") ) {
    output_file = stdout;
    if (verbose) {
      fprintf(stderr, "output file = stdout\n");
    }
  } else {
    // Create the output file:
    switch (output) {
    case KIWI_OUTPUT :  
      strcat(output_filename, ".ktr");
      break;
    case XML_OUTPUT :
      strcat(output_filename, ".xml");
      break;
    case TEXT_OUTPUT :
      strcat(output_filename, ".txt");
      break;
    }

    if (verbose) {
      fprintf(stderr, "output file = %s\n", output_filename);
    }

    if ((output_file=fopen(output_filename, "w")) == NULL) {
      fprintf(stderr, "logto: Error in creating output file %s\n", 
	      output_filename);
      exit(1);
    }
  }


  // Open the log into a pre-recorded trace stream:
  if ((log_filedesc = open(log_filename, O_RDONLY)) < 0) {
      fprintf(stderr, "logto: Error while opening log file %s\n",
	      log_filename);
      exit(1);
  }

  error = posix_trace_open(log_filedesc, &trace_id);
  if (error < 0) {
    fprintf(stderr, "logto: Error: log contents seem to be invalid.\n");
    exit(1);
  }

  // Calculate maxiums (linux "number" amongh others):
  do_calculate_maximums(trace_id, verbose);

  // Generate output file:
  switch (output) {
  case KIWI_OUTPUT :  
    do_kiwi_output (trace_id, output_file, verbose);
    break;
  case XML_OUTPUT :
    do_xml_output  (trace_id, output_file, verbose);
    break;
  case TEXT_OUTPUT :
    do_text_output (trace_id, output_file, verbose);
    break;
  }

 
  // All done. Go.
  fflush(NULL);
  fclose(output_file);
  posix_trace_close(trace_id);

  return 0;
}


